package at.fhooe.mc.magicwand.client.network;

import javax.microedition.io.*;

import at.fhooe.mc.magicwand.client.exceptions.ServerConnectionException;

import java.io.*;

/**
 * @class ClientServerConnection
 * @author Manuela Vogler
 * @brief Handles the conncetion to the server. It opens the port for the
 * socket, an output and an input stream. The class also knows the Magic Wand
 * Protocol in order to create valid messages and to check if received ones are
 * valid.
 */
public class ClientServerConnection {

    /**
     * The socket connection to the server
     */
    private SocketConnection m_con;
    
    /**
     * The output stream used to send messages to the server
     */
    private DataOutputStream m_dos;

    /**
     * The input stream used to receive messages from the server
     */
    private DataInputStream m_dis;

    /**
     * The MwProtocol used to create valid messages and to check if received
     * messages are valid
     */
    private MwProtocol m_prot;

    /**
     * To stop receiving if the user aborts
     */
    private boolean m_receive;

    /**
     * @fn public boolean isM_receive()
     * @brief checks if receiving is active
     * @return if receiving is active
     */
    public boolean isM_receive() {
        return m_receive;
    }

    /**
     * @fn public void setM_receive(boolean _receive)
     * @brief sets receiving active or inactive
     * @param _receive set receiving active or inactive
     */
    public void setM_receive(boolean _receive) {
        m_receive = _receive;
    }

    
    /**
     * @fn public ClientServerConnection(int _port) throws ServerConnectionException
     * @brief Constructs a new ClientServerConnection using the given port.
     * @param _address server address
     * @param _port port for the socket
     * @throws ServerConnectionException if connection can't be established
     */
    public ClientServerConnection(String _address, int _port) throws ServerConnectionException {
        try {
            //Opens socket, data output stream und data input stream
            m_con = (SocketConnection) Connector.open("socket://" + _address + ":" + _port);
            m_dos = m_con.openDataOutputStream();
            m_dis = m_con.openDataInputStream();

            //Initializes new instance of the MwProtocol
            m_prot = new MwProtocol();
        } catch (IOException e) {
            throw new ServerConnectionException("Error while establishing connection to server.");
        }

    }

    /**
     * @fn public String sendAndReceive(String _message) throws ServerConnectionException
     * @brief Sends a message to the server and calls method
     * receiveValidAnswer to check if the answer from the server is valid.
     * @param _message Message that should be sent to the server
     * @return A string with the xml data received from the server
     * @throws ServerConnectionException if an error occured during the sending or receiving 
     * or if the response doesn't conform to the protocol
     */
    public String sendAndReceive(String _message) throws ServerConnectionException {
        //Create a message that conforms to the MwProtocol
        String msg = createMessage(_message);
        String answer;
        try {
            //Send message to server
            m_dos.write(msg.getBytes());
            m_dos.flush();

            //Receive answer
            answer = receiveValidAnswer();
            if (answer.equals("")) {
                throw new ServerConnectionException("Received answer is not valid");
            }
        } catch (IOException e) {
            throw new ServerConnectionException("Error while send and receive");
        }
        return answer;
    }

    /**
     * Creates valid messages. Combines the protocol elements with
     * the reqeust data information.
     * @param _data Data which should be sent to the server
     * @return A valid message (including preambel, length, sepchar
     * and data)
     */
    private String createMessage(String _data) {
        StringBuffer ret = new StringBuffer();
        ret.append(m_prot.getPreambel()); //add preambel
        ret.append(_data.length()); //add length
        ret.append(m_prot.getSepchar()); //add sepchar
        ret.append(_data); //add request data
        return ret.toString();
    }

    /**
     * Receives the answer from the server and uses the MagicWand
     * Protocol to check if it is valid.
     * @return The response (in xml format) or null if an error occured
     * @throws  ServerConnectionException if response can't be received or if response doesn't conform to protocol
     */
    private String receiveValidAnswer() throws ServerConnectionException {
        StringBuffer answer = new StringBuffer();
        char inputChar;
        try {
            while ((inputChar = (char) m_dis.read()) != '\0' && isM_receive()) {
                answer.append(inputChar);
            }
            m_prot.processData(answer.toString());
        } catch (Exception _e) {
            throw new ServerConnectionException("Error while receiving.");
        }

        return m_prot.getXmlAnswer();
    }

    /**
     * @fn public void closeConnection() throws ServerConnectionException
     * @brief Is called when the client wants to close the connection. Sends
     * a special message to the server to signalize the closing of the connection,
     * then closes the input and output streams and the socket connection.
     * @throws ServerConnectionException if byebye message can't be sent or the connection can't be closed
     */
    public void closeConnection() throws ServerConnectionException {
        try {
            m_dos.write(m_prot.getCloseMessage().getBytes()); //send closing message
            m_dos.flush();
            m_dis.close();
            m_dos.close();
            m_con.close();
        } catch (IOException _e) {
            throw new ServerConnectionException("Error while closing connection.");
        }

    }
}

